/**
 * @file cpu_freq_plugin.c
 *
 * @brief Tuning Plugin for CPU frequency tuning
 */
#define _GNU_SOURCE
#include <scorep/rrl_tuning_plugins.h>

#include "fastcpufreq.h"
#include <cpufreq.h>
#include <errno.h>
#include <sched.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <unistd.h>

#define PLUGIN_NAME "cpu_freq"

static cpu_set_t *responible_cpus;
static size_t responible_cpus_size;

static long available_cores;

typedef enum { LOG_VERBOSE, LOG_WARN, LOG_INFO, LOG_DEBUG } log_level;

/**
 * log function.
 *
 * Prints messages depending on the entry in
 *SCOREP_TUNING_CPU_FREQ_PLUGIN_VERBOSE.
 *
 * Currently implemented loge levels are are:
 *
 *	* LOG_VERBOSE
 *	* LOG_WARN
 *	* LOG_INFO
 *	* LOG_DEBUG
 *
 * @param[in] msg_level level of message
 * @param[in] message_fmt printf like message
 * @param[in] ... printf like parameters for the message
 */
void llog(log_level msg_level, const char *message_fmt, ...)
{
    char *level_str;
    log_level level;
    level_str = getenv("SCOREP_TUNING_CPU_FREQ_PLUGIN_VERBOSE");
    if (level_str == NULL)
    {
        level = LOG_WARN;
    }
    else if (strcmp(level_str, "DEBUG") == 0)
    {
        level = LOG_DEBUG;
    }
    else if (strcmp(level_str, "INFO") == 0)
    {
        level = LOG_INFO;
    }
    else if (strcmp(level_str, "WARN") == 0)
    {
        level = LOG_WARN;
    }
    else if (strcmp(level_str, "VERBOSE") == 0)
    {
        level = LOG_VERBOSE;
    }
    else
    {
        level = LOG_WARN;
    }
    if (msg_level <= level)
    {
        char *output_fmt = (char *) malloc(strlen(message_fmt) + strlen(PLUGIN_NAME) + 5);
        strcpy(output_fmt, "[");
        strcpy(output_fmt + 1, PLUGIN_NAME);
        strcpy(output_fmt + 1 + strlen(PLUGIN_NAME), "]");
        strcpy(output_fmt + 2 + strlen(PLUGIN_NAME), message_fmt);
        strcpy(output_fmt + 2 + strlen(PLUGIN_NAME) + strlen(message_fmt), "\n");
        va_list args;
        va_start(args, message_fmt);
        vprintf(output_fmt, args);
        va_end(args);
    }
}

void init_responsilbe_cpus()
{
    int cpu;
    int rt;

    llog(LOG_DEBUG, "init CPUs");

    for (cpu = 0; cpu < available_cores; cpu++)
    {
        if (CPU_ISSET_S(cpu, responible_cpus_size, responible_cpus))
        {
            llog(LOG_DEBUG, "init cpu %d", cpu);
            rt = fcf_init_cpu(cpu);

            if (rt < 0)
            {
                llog(LOG_WARN, "init cpu %d failed:", cpu);
                llog(LOG_WARN, strerror(errno), rt);
            }
        }
    }
}

/**
 * Initialize the plugin
 *
 * counts CPUs and calls get_cpus to get the for this program active ones
 *
 * @retrun 0 at sucess, -1 at failure
 */
int32_t init()
{
    llog(LOG_VERBOSE, "CPU_FREQ tuning plugin: initializing");

    available_cores = sysconf(_SC_NPROCESSORS_ONLN);
    llog(LOG_INFO, "got %u cpus", available_cores);

    // get inital responsible CPUS

    responible_cpus = CPU_ALLOC(available_cores);
    responible_cpus_size = CPU_ALLOC_SIZE(available_cores);

    if (responible_cpus == NULL)
    {
        llog(LOG_WARN, "error CPU_ALLOC");
        return -1;
    }

    CPU_ZERO_S(responible_cpus_size, responible_cpus);
    sched_getaffinity(getpid(), responible_cpus_size, responible_cpus);
    init_responsilbe_cpus();

    return 0;
}

/**
 * Set the frequency
 *
 * Sets the frequency to new_settigns for all CPUs that have a userspace
 * governor.
 * If new_settings is smaller or larger than the minimal or maximal CPU
 * frequency,
 * the frequency is set to the minimal or maximal CPU frequency.
 *
 * @param[in] new_settings new frequency settings for CPUs
 * @param[out] old_setting returns the previous set frequency
 */

static void scorep_set_cpu_freq(int new_settings)
{
    unsigned long freq = (unsigned long) new_settings;

    int cpu;
    int rt;

    llog(LOG_DEBUG, "setting freq to %lu", new_settings);

    for (cpu = 0; cpu < available_cores; cpu++)
    {
        if (CPU_ISSET_S(cpu, responible_cpus_size, responible_cpus))
        {
            llog(LOG_DEBUG, "set cpu %d to %lu", cpu, new_settings);
            rt = fcf_set_frequency(cpu, freq);

            if (rt < 0)
            {
                switch (rt)
                {
                case -1:
                    llog(LOG_WARN,
                        "getting error setting frequency: not initialized (number: %d)",
                        rt);
                    break;
                case -2:
                    llog(LOG_WARN,
                        "getting error setting frequency: not invalid cpu selected "
                        "(number: %d)",
                        rt);
                    break;
                default:
                    llog(LOG_WARN,
                        "getting error setting frequency: unknown error (number: %d)",
                        rt);
                    break;
                }
            }
        }
    }
}

/** gets the freq for !all! cpus, and return the higest.
 *
 */
static int scorep_get_cpu_freq()
{
    int freq = -1;
    int cpu = 0;

    for (cpu = 0; cpu < available_cores; cpu++)
    {
        int cpu_freq = cpufreq_get_freq_kernel(cpu);
        if (freq < cpu_freq)
        {
            freq = cpu_freq;
        }
    }

    return freq;
}

/**
 * Gets the cpu affinity of the CPU Thread and adds it the resposible cpus list
 */

void create_location(RRL_LocationType location_type, uint32_t location_id)
{
    llog(LOG_DEBUG, "create_location for location %u with typ %u ", location_id, location_type);
    if (location_type == RRL_LOCATION_TYPE_CPU_THREAD)
    {
        cpu_set_t *set;
        pid_t tid;
        tid = syscall(SYS_gettid);
        set = CPU_ALLOC(available_cores);
        size_t set_size = CPU_ALLOC_SIZE(available_cores);

        if (set == NULL)
        {
            llog(LOG_WARN, "error during create_location: CPU_ALLOC");
            return;
        }
        CPU_ZERO_S(set_size, set);
        sched_getaffinity(tid, set_size, set);
        CPU_OR_S(responible_cpus_size, responible_cpus, set, responible_cpus);

        CPU_FREE(set);
        llog(LOG_DEBUG, "added new CPU");
        int cpu = 0;
        for (cpu = 0; cpu < available_cores; cpu++)
        {
            if (CPU_ISSET_S(cpu, responible_cpus_size, responible_cpus))
            {
                llog(LOG_DEBUG, "cpu in set: %d ", cpu);
            }
        }
        init_responsilbe_cpus();
    }
}

void delete_location(RRL_LocationType location_type, uint32_t location_id)
{
}

/**
 * finalising the plugin
 */
void fini()
{
    printf("CPU_FREQU tuning plugin: finalizing");
    fcf_finalize();
    CPU_FREE(responible_cpus);
    responible_cpus = 0;
}

/**
 * ScoreP array for plugin definitions
 */
static rrl_tuning_action_info return_values[] = {
    {
        .name = "CPU_FREQ",                              /**< Name of the Plugin*/
        .current_config = &scorep_get_cpu_freq,          /**< get current configuration*/
        .enter_region_set_config = &scorep_set_cpu_freq, /**< function to call on enter event*/
        .exit_region_set_config = &scorep_set_cpu_freq   /**< function to call on exit event*/
    },
    {.name = NULL,
        .current_config = NULL,
        .enter_region_set_config = NULL,
        .exit_region_set_config = NULL}};

/**
 * ScoreP function to get plugin definitions
 *
 * @param return return_values.
 */
rrl_tuning_action_info *get_tuning_info()
{
    return return_values;
}

/**
 * Macro to setup the plugin
 */
RRL_TUNING_PLUGIN_ENTRY(cpu_freq_plugin)
{
    /* Initialize info data (with zero) */
    rrl_tuning_plugin_info info;
    memset(&info, 0, sizeof(rrl_tuning_plugin_info));

    /* Set up */
    info.plugin_version = RRL_TUNING_PLUGIN_VERSION;
    info.initialize = init;
    info.get_tuning_info = get_tuning_info;
    info.finalize = fini;
    info.create_location = create_location;
    info.delete_location = delete_location;
    return info;
}
