/**
 * @file sample_plugin.c
 *
 * @brief Description of a sample plugin
 */

#define _GNU_SOURCE
#include <scorep/rrl_tuning_plugins.h>

#include <errno.h>
#include <sched.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <unistd.h>
#include <x86_adapt.h>

static x86_adapt_device_type adapt_type;
static int available_devices;
static int *devices;
static int ufm_num;
static int ufm_num_min;
static int ufm_num_max;

static char *env_string;

static int cores_per_die;
static int *socet_owned;
static int check_fully_occupied = 0;

static long available_cores;
static cpu_set_t *responible_cpus;
static size_t responible_cpus_size;

#define UNCORE_CURRENT_RATIO "Intel_UNCORE_CURRENT_RATIO"
#define UNCORE_MIN_RATIO "Intel_UNCORE_MIN_RATIO"
#define UNCORE_MAX_RATIO "Intel_UNCORE_MAX_RATIO"
#define PLUGIN_NAME "UNCORE_FREQ_TP"

typedef enum { LOG_VERBOSE, LOG_WARN, LOG_INFO, LOG_DEBUG } log_level;

/**
 * log function.
 *
 * Prints messages depending on the entry in SCOREP_TUNING_CPU_FREQ_PLUGIN_VERBOSE.
 *
 * Currently implemented loge levels are are:
 *
 *	* LOG_VERBOSE
 *	* LOG_WARN
 *	* LOG_INFO
 *	* LOG_DEBUG
 *
 * @param[in] msg_level level of message
 * @param[in] message_fmt printf like message
 * @param[in] ... printf like parameters for the message
 */
void llog(log_level msg_level, const char *message_fmt, ...)
{
    char *level_str;
    log_level level;
    level_str = getenv("SCOREP_TUNING_UNCORE_FREQ_PLUGIN_VERBOSE");
    if (level_str == NULL)
    {
        level = LOG_WARN;
    }
    else if (strcmp(level_str, "DEBUG") == 0)
    {
        level = LOG_DEBUG;
    }
    else if (strcmp(level_str, "INFO") == 0)
    {
        level = LOG_INFO;
    }
    else if (strcmp(level_str, "WARN") == 0)
    {
        level = LOG_WARN;
    }
    else if (strcmp(level_str, "VERBOSE") == 0)
    {
        level = LOG_VERBOSE;
    }
    else
    {
        level = LOG_WARN;
    }
    if (msg_level <= level)
    {
        char *output_fmt = (char *) malloc(strlen(message_fmt) + strlen(PLUGIN_NAME) + 5);
        strcpy(output_fmt, "[");
        strcpy(output_fmt + 1, PLUGIN_NAME);
        strcpy(output_fmt + 1 + strlen(PLUGIN_NAME), "]");
        strcpy(output_fmt + 2 + strlen(PLUGIN_NAME), message_fmt);
        strcpy(output_fmt + 2 + strlen(PLUGIN_NAME) + strlen(message_fmt), "\n");
        va_list args;
        va_start(args, message_fmt);
        vprintf(output_fmt, args);
        va_end(args);
    }
}

/**
 * Initialize the plugin
 *
 * looks for all CPUs that have a userspace governor.
 *
 * @return 0 at sucess
 */
int32_t init()
{
    // some initialisation
    llog(LOG_VERBOSE, "UNCORE_FREQ tuning plugin: initializing\n");
    if (x86_adapt_init())
    {
        llog(LOG_WARN, "Could not initialize x86_adapt library");
    }
    env_string = getenv("CHECK_IF_NODE_FULLY_OCCUPIED");
    if (env_string == NULL)
        check_fully_occupied = 1;
    else
    {
        check_fully_occupied = atoi(env_string);
        if (check_fully_occupied == 1)
            llog(LOG_INFO,
                "CHECK_IF_NODE_FULLY_OCCUPIED is enabled. Uncore frequency only set "
                "when all the cores on the node are occupied");
        else if (check_fully_occupied != 0)
        {
            llog(LOG_WARN,
                "Could not parse the value provided. CHECK_IF_NODE_FULLY_OCCUPIED "
                "should be set to 0 or 1");
            check_fully_occupied = 0;
        }
    }

    adapt_type = X86_ADAPT_DIE;
    available_devices = x86_adapt_get_nr_avaible_devices(adapt_type);
    devices = calloc(available_devices, sizeof(int));
    int node = 0;
    for (node = 0; node < available_devices; node++)
    {
        devices[node] = x86_adapt_get_device(adapt_type, node);
    }

    available_cores = sysconf(_SC_NPROCESSORS_ONLN);

    llog(LOG_DEBUG, "got %u devices", available_devices);
    llog(LOG_DEBUG, "got %u cpus", available_cores);

    cores_per_die = available_cores / available_devices;
    llog(LOG_INFO, "assuming %d cores per device (processor die)", cores_per_die);

    ufm_num = x86_adapt_lookup_ci_name(adapt_type, UNCORE_CURRENT_RATIO);
    if (ufm_num < 0)
    {
        llog(LOG_WARN, "[init]could not find item %s", UNCORE_CURRENT_RATIO);
    }
    llog(LOG_DEBUG, "[init]found item %s with number %d", UNCORE_CURRENT_RATIO, ufm_num);

    ufm_num_min = x86_adapt_lookup_ci_name(adapt_type, UNCORE_MIN_RATIO);
    if (ufm_num_min < 0)
    {
        llog(LOG_WARN, "[init]could not find item %s", UNCORE_MIN_RATIO);
    }
    llog(LOG_DEBUG, "[init]found item %s with number %d", UNCORE_MIN_RATIO, ufm_num_min);

    ufm_num_max = x86_adapt_lookup_ci_name(adapt_type, UNCORE_MAX_RATIO);
    if (ufm_num_max < 0)
    {
        llog(LOG_WARN, "[init]could not find item %s", UNCORE_MAX_RATIO);
    }
    llog(LOG_DEBUG, "[init]found item %s with number %d", UNCORE_MAX_RATIO, ufm_num_max);

    // get inital responsible CPUS

    responible_cpus = CPU_ALLOC(available_cores);
    responible_cpus_size = CPU_ALLOC_SIZE(available_cores);

    if (responible_cpus == NULL)
    {
        llog(LOG_WARN, "error CPU_ALLOC");
        return -1;
    }

    CPU_ZERO_S(responible_cpus_size, responible_cpus);

    pid_t tid;
    tid = syscall(SYS_gettid);
    sched_getaffinity(tid, responible_cpus_size, responible_cpus);

    socet_owned = calloc(sizeof(int), available_devices);

    return 0;
}

void create_location(RRL_LocationType location_type, uint32_t location_id)
{
    llog(LOG_DEBUG, "create_location for location %u with typ %u ", location_id, location_type);
    if (location_type == RRL_LOCATION_TYPE_CPU_THREAD)
    {
        cpu_set_t *set;
        pid_t tid;
        size_t set_size = CPU_ALLOC_SIZE(available_cores);

        tid = syscall(SYS_gettid);
        set = CPU_ALLOC(available_cores);

        if (set == NULL)
        {
            llog(LOG_WARN, "error during create_location: CPU_ALLOC");
            return;
        }

        CPU_ZERO_S(set_size, set);
        sched_getaffinity(tid, set_size, set);
        CPU_OR_S(responible_cpus_size, responible_cpus, set, responible_cpus);

        CPU_FREE(set);
        llog(LOG_DEBUG, "added new CPU");
        int cpu = 0;
        for (cpu = 0; cpu < available_cores; cpu++)
        {
            if (CPU_ISSET_S(cpu, responible_cpus_size, responible_cpus))
            {
                llog(LOG_DEBUG, "cpu in set: %d ", cpu);
            }
        }

        /* check which cpus we own on which die:
         *
         */
        int node;
        for (node = 0; node < available_devices; node++)
        {

            int owned_cpus_on_socet = 0;

            for (cpu = node * cores_per_die; cpu < cores_per_die * (node + 1); cpu++)
            {
                if (CPU_ISSET_S(cpu, responible_cpus_size, responible_cpus))
                {
                    owned_cpus_on_socet++;
                }
            }
            if (owned_cpus_on_socet < cores_per_die)
            {
                socet_owned[node] = 0;
                llog(LOG_DEBUG,
                    "got %d out of %d cores on die %d. Won't start tuning the uncore",
                    owned_cpus_on_socet,
                    cores_per_die,
                    node);
            }
            else if (owned_cpus_on_socet == cores_per_die)
            {
                socet_owned[node] = 1;
                llog(LOG_INFO,
                    "got %d out of %d cores on die %d. Start tuning the uncore",
                    owned_cpus_on_socet,
                    cores_per_die,
                    node);
            }
            else if (owned_cpus_on_socet > cores_per_die)
            {
                socet_owned[node] = 0;
                llog(LOG_WARN,
                    "got %d out of %d cores on die %d. Won't start tuning the uncore. \n"
                    "YOU GOT MORE CORES ASSOCIATED TO YOUR DIE, THAN THERE ARE CORES ON YOUR DIE\n"
                    "YOU ARE EITHER A MAGE OR SOMETHING WENT REALLY WRONG :-)",
                    owned_cpus_on_socet,
                    cores_per_die,
                    node);
            }
            else
            {
                llog(LOG_WARN,
                    "This should never happen XD",
                    owned_cpus_on_socet,
                    cores_per_die,
                    node);
            }
        }
    }
}

void delete_location(RRL_LocationType location_type, uint32_t location_id)
{
}

/**
 * finalising the plugin
 */
void fini()
{
    int node = 0;
    for (node = 0; node < available_devices; node++)
    {
        x86_adapt_put_device(adapt_type, node);
    }
    x86_adapt_finalize();
    // some finalisation
    printf("UNCORE_FREQ tuning plugin: finalizing\n");
}

static void scorep_set_uncore_freq(int new_settings)
{
    unsigned long new_settings_ = (unsigned long) new_settings;
    int node = 0;
    int rt1, rt2;

    for (node = 0; node < available_devices; node++)
    {

        /* Set the uncore freq
         *
         */
        if (!check_fully_occupied || socet_owned[node])
        {

            if ((rt1 = x86_adapt_set_setting(devices[node], ufm_num_min, new_settings_)) != 8 ||
                (rt2 = x86_adapt_set_setting(devices[node], ufm_num_max, new_settings_)) != 8)
            {
                llog(LOG_WARN, "Could not set uncore frequency for node %d\n", node);
            }

            llog(LOG_INFO,
                "setting uncore frequency to %lu (node: %d)\n",
                new_settings_,
                node);
        }
    }
}

/**
 * returns the average of the uncore frequency
 *
 * looks for all CPU's frequencies and makes their average
 *
 * @return average
 */
static int scorep_get_uncore_freq()
{
    unsigned long int unc_freq = 0;
    unsigned long int average = 0;
    int count = 0;
    int node = 0;
    int rt;

    for (node = 0; node < available_devices; node++)
    {

        if ((rt = x86_adapt_get_setting(devices[node], ufm_num, &unc_freq)) != 8)
        {
            llog(LOG_WARN, "Could not get uncore frequency for node %d\n", node);
        }
        else
        {
            ++count;
            average += unc_freq;
        }
    }

    if (count > 1)
    {
        return (int) average / count;
    }
    else
    {
        return (int) average;
    }
}

/**
 * ScoreP array for plugin definitions
 */
static rrl_tuning_action_info return_values[] = {
    {
        .name = "UNCORE_FREQ",
        .current_config = &scorep_get_uncore_freq,
        .enter_region_set_config = &scorep_set_uncore_freq,
        .exit_region_set_config = &scorep_set_uncore_freq,
    },
    {
        .name = NULL,
        .current_config = NULL,
        .enter_region_set_config = NULL,
        .exit_region_set_config = NULL,
    }};

/**
 * ScoreP function to get plugin definitions
 *
 * @param return return_values.
 */
rrl_tuning_action_info *get_tuning_info()
{
    return return_values;
}

/**
 * Macro to setup the plugin
 */
RRL_TUNING_PLUGIN_ENTRY(uncore_freq_plugin)
{
    /* Initialize info data (with zero) */
    rrl_tuning_plugin_info info;
    memset(&info, 0, sizeof(rrl_tuning_plugin_info));

    /* Set up */
    info.plugin_version = RRL_TUNING_PLUGIN_VERSION;
    info.initialize = init;
    info.get_tuning_info = get_tuning_info;
    info.finalize = fini;
    info.create_location = create_location;
    info.delete_location = delete_location;
    return info;
}
